/**
 * \file snow.c
 * \brief Snow Effect Demo
 *
 * lionel_debroux@yahoo.fr (TICT) 2010-2011
 *
 * Port to the Nspire (Ndless pre-2.0) of the TI-68k version made in 2000-2002
 * by Thomas Nussbaumer & Lionel Debroux, with two modifications:
 * - this version uses a LFSR (pseudo-RNG) for emulating the flame decay
 *   effect, instead of a larger and slower RNG;
 * - this version provides a way to adjust contrast during execution, in
 *   order to make it easier to see moving objects on that slow screen
 *   of terrible quality.
 *
 * This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details.
 */

#include <os.h>
#include <libndls.h>
#include <stdint.h>


#define WAIT(n, name) \
  asm volatile ( \
    "ldr r0, ="n"\n" \
    "_loop_"name":\n" \
    "subs r0, r0, #1\n" \
    "bne _loop_"name"\n" \
    : : : "r0" );


//! Generator for 32-bit numbers.
#define LFSR_XOR_MASK (0xA3000000)

#define BITS_PER_PIXEL  (4)
#define BITS_PER_BYTE   (8)
#define SCREEN_BYTES_PER_LINE (SCREEN_WIDTH * BITS_PER_PIXEL / BITS_PER_BYTE)

//! Maximum number of snow particles to handle
#define NR_PARTICLES  (1500)
//! Indicates an unused particle entry
#define FREE_ENTRY    (0xFFFF)
//! Clamp the number of particles per line.
#define NR_PARTICLES_PER_LINE (4)


//! Structure which stores a snow particle's position
typedef struct {
    uint16_t offset;  // byte offset of particle (if this is equal
                      // FREE_ENTRY then this entry is free to re-use)
    uint8_t  mask;    // bit mask of pixel
    // 1 byte wasted...
} particle;

//! Array of particle positions
particle * positions;
//! Number of snow particles
uint32_t count = 0;
//! Holds the LFSR value
uint32_t seq;


/**
 * Fills particles array
 *
 * NOTE: an optimization hint is to define "static" internal routines which
 *       are used just once or twice.
 *       This enables the optimizer to completely inline them, it if can
 *       determine that it will result in better code.
 */
static void FillParticlesArray(void) {
    uint32_t i;
    uint32_t j;
    uint32_t k; // Limit the number of snow particles per line.
    uint16_t offset = 0;
    uint8_t  mask;

    for (i = count, j = 0, k = 0; i < NR_PARTICLES && k < NR_PARTICLES_PER_LINE; i++, k++) {
        // Search for next free entry in particles array
        while (j < NR_PARTICLES && positions[j].offset != FREE_ENTRY) {
            j++;
        }
        if (j >= NR_PARTICLES) {
            return;
        }

        // Select a random pixel and calculate its byte offset and its bitmask
        // This code assumes 4 bits per pixel, white <-> 1111 and black <-> 0000.
        if (seq & 1) {
            mask = 0x0F;
            seq = (seq >> 1) ^ LFSR_XOR_MASK;
        }
        else {
            mask = 0xF0;
            seq = seq >> 1;
        }

        offset = seq % SCREEN_BYTES_PER_LINE;

        count++;
        positions[j].offset = offset;
        positions[j].mask   = mask;
    }
}

/**
 * Process particles in particles array
 *
 * NOTE: an optimization hint is to define "static" internal routines which
 *       are used just once or twice.
 *       This enables the optimizer to completely inline them, it if can
 *       determine that it will result in better code.
 */
static void ProcessParticlesArray(void) {
    uint32_t i;
    particle * ptr = &positions[0];

    for (i = 0; i < NR_PARTICLES; i++) {
        // If this entry is not marked as free, check if the pixel just below it is set.
        // If there is no pixel just below it, move the current pixel one line down
        // This code assumes 4 bits per pixel, white <-> 1111 and black <-> 0000.
        if (ptr->offset != FREE_ENTRY) {
            if ((*((uint8_t *)SCREEN_BASE_ADDRESS + ptr->offset + SCREEN_BYTES_PER_LINE) & ptr->mask) != 0) {
                *((uint8_t *)SCREEN_BASE_ADDRESS + ptr->offset) |= (ptr->mask);
                ptr->offset += SCREEN_BYTES_PER_LINE;
                // If the entry falls below the bottom of the screen, mark it as free.
                if (ptr->offset >= SCREEN_BYTES_SIZE) {
                    ptr->offset = FREE_ENTRY;
                    count--;
                }
                else {
                    *((uint8_t *)SCREEN_BASE_ADDRESS + ptr->offset) &= ~(ptr->mask);
                }
            }
            else {
                // If there is a pixel exactly one line below the current one,
                // mark the current particle entry as free, because it cannot "fall" down
                ptr->offset = FREE_ENTRY;
                count--;
            }
        }

        ptr++;
    }
}


/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
int main(void) {
    uint32_t initial_contrast = *((uint32_t *)0x900F0020);

    seq = 1234567891;

    show_msgbox(
        "snow - Snow effect demo",
        "------------------------\n"
        "    Snow effect demo\n"
        "    © 2000-2010 TICT\n"
        "------------------------\n"
        "BKSPC - Reset the screen\n"
        "RIGHT - Increase contrast\n"
        "LEFT  - Decrease contrast\n"
        "ESC   - Exit"
    );

    positions = malloc(sizeof(particle)*NR_PARTICLES);
    if (positions != NULL) {
        uint8_t blackline[SCREEN_BYTES_PER_LINE];

        memset(blackline, 0x00, SCREEN_BYTES_PER_LINE);

begin:
        count = 0;

        // Mark all entries in the particles array as free.
        // FREE_ENTRY is 0xFFFF, so we can just use memset here.
        memset(positions, 0xFFFF, sizeof(particle)*NR_PARTICLES);

        // Setup the screen: one black line at the bottom, white in-between.
        memset(SCREEN_BASE_ADDRESS + SCREEN_BYTES_PER_LINE, 0xFF, SCREEN_BYTES_SIZE - 2 * SCREEN_BYTES_PER_LINE);
        memcpy(SCREEN_BASE_ADDRESS + SCREEN_BYTES_SIZE - SCREEN_BYTES_PER_LINE, blackline, SCREEN_BYTES_PER_LINE);

        //---------------------------------------------
        // main loop
        //---------------------------------------------
        while (!isKeyPressed(KEY_NSPIRE_ESC)) {
            // Redraw full black line at the top.
            memcpy(SCREEN_BASE_ADDRESS, blackline, SCREEN_BYTES_PER_LINE);

            if (count < NR_PARTICLES-1) {
                FillParticlesArray();
            }
            ProcessParticlesArray();

            WAIT("0x3FFFF", "wait");

            // Check if the screen is (almost) completely black...
            if (!memcmp(SCREEN_BASE_ADDRESS + SCREEN_BYTES_PER_LINE, blackline, SCREEN_BYTES_PER_LINE)) {
                goto begin;
            }

            if (isKeyPressed(KEY_NSPIRE_CLEAR)) {
                goto begin;
            }
            else if (isKeyPressed(KEY_NSPIRE_RIGHT) && (*((uint32_t *)0x900F0020) < 0x95)) {
                // Increase contrast.
                *(uint32_t *)0x900F0020 += 1;
            }
            else if (isKeyPressed(KEY_NSPIRE_LEFT) && (*((uint32_t *)0x900F0020) > 0x6B)) {
                // Decrease contrast.
                *(uint32_t *)0x900F0020 -= 1;
            }
        }

        // Cleanup and bye =8-)
        // Restore contrast
        *((uint32_t *)0x900F0020) = initial_contrast;
        free(positions);
    }

    return 0;
}
